
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#ifndef INCLUDE_RCF_RCFSESSION_HPP
#define INCLUDE_RCF_RCFSESSION_HPP

#include <vector>

#include <boost/any.hpp>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>

#include <RCF/Export.hpp>
#include <RCF/MethodInvocation.hpp>
#include <RCF/SerializationProtocol.hpp>
#include <RCF/ServerTransport.hpp>
#include <RCF/StubEntry.hpp>

namespace RCF {

    class Filter;

    typedef boost::shared_ptr<Filter> FilterPtr;

    class RcfSession;

    typedef boost::shared_ptr<RcfSession> RcfSessionPtr;

    class RCF_EXPORT RcfSession : public I_Session
    {
    public:
        RcfSession();
        ~RcfSession();

        typedef boost::function1<void, RcfSession&> OnWriteCompletedCallback;
        typedef boost::function1<void, RcfSession&> OnWriteInitiatedCallback;
        typedef boost::function1<void, RcfSession&> OnDestroyCallback;

        //*******************************
        // callback tables - synchronized

        // may well be called on a different thread than the one that executed the remote call
        void addOnWriteCompletedCallback(
            const OnWriteCompletedCallback &onWriteCompletedCallback);

        void extractOnWriteCompletedCallbacks(
            std::vector<OnWriteCompletedCallback> &onWriteCompletedCallbacks);

        // called on same thread that executed the remote call
        void addOnWriteInitiatedCallback(
            const OnWriteInitiatedCallback &onWriteInitiatedCallback);

        void extractOnWriteInitiatedCallbacks(
            std::vector<OnWriteInitiatedCallback> &onWriteInitiatedCallbacks);

        void setOnDestroyCallback(OnDestroyCallback onDestroyCallback);

        //*******************************

        const RCF::I_RemoteAddress &
                        getRemoteAddress();

        void            disconnect();

        bool            hasDefaultServerStub();
        StubEntryPtr    getDefaultStubEntryPtr();
        void            setDefaultStubEntryPtr(const StubEntryPtr &stubEntryPtr);
        void            setCachedStubEntryPtr(const StubEntryPtr &stubEntryPtr);

        /// Enables pointer tracking for outbound SF serialization on this session.
        /// \parameter Whether to enable or not.
        void            enableSfSerializationPointerTracking(bool enable);

        int             getRcfRuntimeVersion();
        void            setRcfRuntimeVersion(int version);

        void            setUserData(boost::any userData);
        boost::any      getUserData();

        void            getMessageFilters(std::vector<FilterPtr> &filters);
        void            getTransportFilters(std::vector<FilterPtr> &filters);

        void            lockTransportFilters();
        void            unlockTransportFilters();
        bool            transportFiltersLocked();

    private:

        StubEntryPtr                            mDefaultStubEntryPtr;
        StubEntryPtr                            mCachedStubEntryPtr;

        Mutex                                   mMutex;
        std::vector<OnWriteCompletedCallback>   mOnWriteCompletedCallbacks;
        std::vector<OnWriteInitiatedCallback>   mOnWriteInitiatedCallbacks;
        OnDestroyCallback                       mOnDestroyCallback;

        int                                     mRcfRuntimeVersion;

        boost::any                              mUserData;

        bool                                    mTransportFiltersLocked;

    public:
        Mutex                                   mStopCallInProgressMutex;
        bool                                    mStopCallInProgress;

        // protocol-opaque serialization
        SerializationProtocolIn                 mIn;
        SerializationProtocolOut                mOut;

        // message filters
        std::vector<FilterPtr>                  mFilters;
        bool                                    mFiltered;

        MethodInvocationRequest                 mRequest;

    };

} // namespace RCF

#endif // ! INCLUDE_RCF_RCFSESSION_HPP
