
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#ifndef INCLUDE_RCF_SSPIFILTER_HPP
#define INCLUDE_RCF_SSPIFILTER_HPP

#include <memory>

#include <boost/enable_shared_from_this.hpp>
#include <boost/shared_ptr.hpp>

#include <RCF/AsyncFilter.hpp>
#include <RCF/Export.hpp>
#include <RCF/RecursionLimiter.hpp>
#include <RCF/Tools.hpp>

#include <RCF/util/Tchar.hpp>

#ifndef SECURITY_WIN32
#define SECURITY_WIN32
#endif

#include <security.h>
#include <tchar.h>

namespace RCF {

    typedef util::tstring tstring;

    RCF_EXPORT tstring getMyUserName();
    RCF_EXPORT tstring getMyDomain();
    RCF_EXPORT tstring getMyMachineName();

    class SspiFilterBase;

    typedef boost::shared_ptr<SspiFilterBase> SspiFilterBasePtr;

    class RCF_EXPORT SspiImpersonator
    {
    public:
        SspiImpersonator(SspiFilterBasePtr sspiFilterPtr);
        ~SspiImpersonator();

        bool impersonate();
        void revertToSelf() const;
    private:
        SspiFilterBasePtr mSspiFilterPtr;
    };

    //static const ULONG DefaultSspiContextRequirements = 2079;
    static const ULONG DefaultSspiContextRequirements =
        ISC_REQ_REPLAY_DETECT   |
        ISC_REQ_SEQUENCE_DETECT |
        ISC_REQ_CONFIDENTIALITY |
        ISC_REQ_INTEGRITY       |
        ISC_REQ_DELEGATE        |
        ISC_REQ_MUTUAL_AUTH;

    class RCF_EXPORT SspiFilterBase :
        public IdentityFilter, // TODO: replace with Filter
        public boost::enable_shared_from_this<SspiFilterBase>
    {
    public:

        ~SspiFilterBase();

        enum QualityOfProtection
        {
            None,
            Encryption,
            Integrity
        };

        QualityOfProtection getQop();

        typedef SspiImpersonator Impersonator;

    protected:

        friend class SspiImpersonator;

        SspiFilterBase(
            const tstring &         packageName,
            const tstring &         packageList,
            bool                    server = false);

        SspiFilterBase(
            const tstring &         target,
            QualityOfProtection     qop,
            ULONG                   contextRequirements,
            const tstring &         packageName,
            const tstring &         packageList,
            bool                    server = false);

        SspiFilterBase(
            const tstring &         userName,
            const tstring &         password,
            const tstring &         domain,
            const tstring &         target,
            QualityOfProtection     qop,
            ULONG                   contextRequirements,
            const tstring &         packageName,
            const tstring &         packageList,
            bool                    server = false);

        enum Event
        {
            ReadIssued,
            WriteIssued,
            ReadCompleted,
            WriteCompleted
        };

        enum ContextState
        {
            AuthContinue,
            AuthOk,
            AuthOkAck,
            AuthFailed
        };

        enum State
        {
            Ready,
            Reading,
            Writing
        };

        PSecurityFunctionTable 
                        getSecurityFunctionTable() const;

        void            acquireCredentials(
                            const tstring &userName = RCF_T(""),
                            const tstring &password = RCF_T(""),
                            const tstring &domain = RCF_T(""));

        void            freeCredentials();

        void            init();
        void            deinit();
        void            reset();

        void            read(
                            const ByteBuffer &byteBuffer, 
                            std::size_t bytesRequested);

        void            write(const std::vector<ByteBuffer> &byteBuffers);

        void            onReadCompleted(const ByteBuffer &byteBuffer, int error);
        void            onWriteCompleted(std::size_t bytesTransferred, int error);

        void            handleEvent(Event event);
        void            readBuffer();
        void            writeBuffer();
        void            encryptWriteBuffer();
        void            decryptReadBuffer();
        bool            completeReadBlock();
        bool            completeWriteBlock();
        bool            completeBlock();
        void            resumeUserIo();
        void            resizeReadBuffer(std::size_t newSize);
        void            resizeWriteBuffer(std::size_t newSize);

        virtual void    handleHandshakeEvent() = 0;

    protected:

        const tstring                           mPackageName;
        const tstring                           mPackageList;
        const tstring                           mTarget;
        QualityOfProtection                     mQop;
        ULONG                                   mContextRequirements;

        bool                                    mHaveContext;
        bool                                    mHaveCredentials;
        bool                                    mImplicitCredentials;
        SecPkgInfo                              mPkgInfo;
        CtxtHandle                              mContext;
        CredHandle                              mCredentials;

        ContextState                            mContextState;
        State                                   mPreState;
        State                                   mPostState;
        Event                                   mEvent;
        const bool                              mServer;

        ByteBuffer                              mReadByteBufferOrig;
        ByteBuffer                              mWriteByteBufferOrig;
        std::size_t                             mBytesRequestedOrig;

        ByteBuffer                              mReadByteBuffer;
        boost::shared_ptr<std::vector<char> >   mReadBufferVectorPtr;
        char *                                  mReadBuffer;
        std::size_t                             mReadBufferPos;
        std::size_t                             mReadBufferLen;

        ByteBuffer                              mWriteByteBuffer;
        boost::shared_ptr<std::vector<char> >   mWriteBufferVectorPtr;
        char *                                  mWriteBuffer;
        std::size_t                             mWriteBufferPos;
        std::size_t                             mWriteBufferLen;

        std::vector<ByteBuffer>                 mByteBuffers;
        ByteBuffer                              mTempByteBuffer;

    private:
        bool                                    mLimitRecursion;
        RecursionState<ByteBuffer, int>         mRecursionStateRead;
        RecursionState<std::size_t, int>        mRecursionStateWrite;

        void onReadCompleted_(const ByteBuffer &byteBuffer, int error);
        void onWriteCompleted_(std::size_t bytesTransferred, int error);
    };

    // server filters

    class RCF_EXPORT SspiServerFilter : public SspiFilterBase
    {
    public:
        SspiServerFilter(
            const tstring &packageName, 
            const tstring &packageList);

    private:
        bool doHandshake();
        void handleHandshakeEvent();
    };

    class RCF_EXPORT SspiNtlmServerFilter : public SspiServerFilter
    {
    public:
        SspiNtlmServerFilter();
        const FilterDescription & getFilterDescription() const;
        static const FilterDescription & sGetFilterDescription();

        // TODO: should be private
        static const FilterDescription *spFilterDescription;
    };

    class RCF_EXPORT SspiKerberosServerFilter : public SspiServerFilter
    {
    public:
        SspiKerberosServerFilter();
        const FilterDescription & getFilterDescription() const;
        static const FilterDescription & sGetFilterDescription();

        // TODO: should be private
        static const FilterDescription *spFilterDescription;
    };

    class RCF_EXPORT SspiNegotiateServerFilter : public SspiServerFilter
    {
    public:
        SspiNegotiateServerFilter(const tstring &packageList);
        const FilterDescription & getFilterDescription() const;
        static const FilterDescription & sGetFilterDescription();

        // TODO: should be private
        static const FilterDescription *spFilterDescription;
    };

    // filter factories

    class RCF_EXPORT SspiNtlmFilterFactory : public FilterFactory
    {
    public:
        FilterPtr createFilter();
        const FilterDescription & getFilterDescription();
    };

    class RCF_EXPORT SspiKerberosFilterFactory : public FilterFactory
    {
    public:
        FilterPtr createFilter();
        const FilterDescription & getFilterDescription();
    };

    class RCF_EXPORT SspiNegotiateFilterFactory : public FilterFactory
    {
    public:
        SspiNegotiateFilterFactory(const tstring &packageList = RCF_T("Kerberos, NTLM"));
        FilterPtr createFilter();
        const FilterDescription & getFilterDescription();
    private:
        tstring mPackageList;
    };

    // client filters

    class RCF_EXPORT SspiClientFilter : public SspiFilterBase
    {
    public:
        SspiClientFilter(
            const tstring &         userName,
            const tstring &         password,
            const tstring &         domain,
            const tstring &         targetName,
            QualityOfProtection     qop,
            ULONG                   contextRequirements,
            const tstring &         packageName,
            const tstring &         packageList) :
                SspiFilterBase(
                    userName, 
                    password, 
                    domain,
                    targetName, 
                    qop, 
                    contextRequirements, 
                    packageName, 
                    packageList)
        {}

        SspiClientFilter(
            const tstring &         targetName,
            QualityOfProtection     qop,
            ULONG                   contextRequirements,
            const tstring &         packageName,
            const tstring &         packageList) :
                SspiFilterBase(
                    targetName, 
                    qop, 
                    contextRequirements, 
                    packageName, 
                    packageList)
        {}

    private:
        bool doHandshake();
        void handleHandshakeEvent();
    };

    class RCF_EXPORT SspiNtlmClientFilter : public SspiClientFilter
    {
    public:
        SspiNtlmClientFilter(
            const tstring &         userName,
            const tstring &         password,
            const tstring &         domain,
            QualityOfProtection     qop = SspiFilterBase::Encryption,
            ULONG                   contextRequirements 
                                        = DefaultSspiContextRequirements);

        SspiNtlmClientFilter(
            QualityOfProtection     qop = SspiFilterBase::Encryption,
            ULONG                   contextRequirements 
                                        = DefaultSspiContextRequirements);

        const FilterDescription &           getFilterDescription() const;
        static const FilterDescription &    sGetFilterDescription();
    };

    class RCF_EXPORT SspiKerberosClientFilter : public SspiClientFilter
    {
    public:
        SspiKerberosClientFilter(
            const tstring &         userName,
            const tstring &         password,
            const tstring &         domain,
            const tstring &         targetName,
            QualityOfProtection     qop = SspiFilterBase::Encryption,
            ULONG                   contextRequirements 
                                        = DefaultSspiContextRequirements);

        SspiKerberosClientFilter(
            const tstring &         targetName,
            QualityOfProtection     qop = SspiFilterBase::Encryption,
            ULONG                   contextRequirements 
                                        = DefaultSspiContextRequirements);

        const FilterDescription &           getFilterDescription() const;
        static const FilterDescription &    sGetFilterDescription();
    };

    class RCF_EXPORT SspiNegotiateClientFilter : public SspiClientFilter
    {
    public:
        SspiNegotiateClientFilter(
            const tstring &         userName,
            const tstring &         password,
            const tstring &         domain,
            const tstring &         targetName,
            QualityOfProtection     qop = SspiFilterBase::None,
            ULONG                   contextRequirements 
                                        = DefaultSspiContextRequirements);

        SspiNegotiateClientFilter(
            const tstring &         targetName,
            QualityOfProtection     qop = SspiFilterBase::Encryption,
            ULONG                   contextRequirements 
                                        = DefaultSspiContextRequirements);

        const FilterDescription &           getFilterDescription() const;
        static const FilterDescription &    sGetFilterDescription();
    };

    typedef SspiNtlmClientFilter        SspiNtlmFilter;
    typedef SspiKerberosClientFilter    SspiKerberosFilter;
    typedef SspiNegotiateClientFilter   SspiNegotiateFilter;

} // namespace RCF

#endif // ! INCLUDE_RCF_SSPIFILTER_HPP
