
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#ifndef INCLUDE_RCF_UDPSERVERTRANSPORT_HPP
#define INCLUDE_RCF_UDPSERVERTRANSPORT_HPP

#include <string>
#include <vector>

#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>

#include <RCF/Export.hpp>
#include <RCF/IpAddress.hpp>
#include <RCF/Service.hpp>
#include <RCF/ServerTransport.hpp>
#include <RCF/IpServerTransport.hpp>
#include <RCF/ThreadLibrary.hpp>

namespace RCF {
   
    class UdpServerTransport;
    class UdpSessionState;
    class UdpProactor;

    typedef boost::shared_ptr<UdpServerTransport> UdpServerTransportPtr;
    typedef boost::shared_ptr<UdpSessionState> UdpSessionStatePtr;

    class RCF_EXPORT UdpProactor : public I_Proactor
    {
    public:
        UdpProactor(
            UdpServerTransport &        transport,
            const UdpSessionStatePtr &  sessionStatePtr);

        I_ServerTransport & getServerTransport();
        UdpSessionStatePtr  getSessionStatePtr() const;

        const I_RemoteAddress &
                            getRemoteAddress();

        void                setTransportFilters(
                                const std::vector<FilterPtr> &filters);

        void                getTransportFilters(
                                std::vector<FilterPtr> &filters);

        ByteBuffer          getReadByteBuffer();
        void                postRead();
        
        void                postWrite(
                                const std::vector<ByteBuffer> &byteBuffers);

        void                postClose();

        int                    getNativeHandle() const;

    private:
        friend class UdpServerTransport;
        UdpServerTransport &    mTransport;
        UdpSessionStatePtr      mSessionStatePtr;
    };

    class UdpSessionState
    {
    public:
        UdpSessionState();

        boost::shared_ptr< std::vector<char> > mReadVecPtr;
        boost::shared_ptr< std::vector<char> > mWriteVecPtr;

        const I_RemoteAddress &getRemoteAddress() const;
        IpAddress remoteAddress;
    };

    class RCF_EXPORT UdpServerTransport :
        public I_ServerTransport,
        public I_IpServerTransport,
        public I_Service,
        boost::noncopyable
    {
    public:

        UdpServerTransport(int port = 0);

        UdpServerTransport(
            const std::string &        networkInterface, 
            int                        port = 0, 
            const std::string &        multicastIp = "");

        ServerTransportPtr 
                    clone();

        I_SessionManager &
                    getSessionManager();

        void        setSessionManager(I_SessionManager &sessionManager);
        void        setPort(int port);
        int         getPort() const;
        void        open();
        void        close();
        void        cycle(int timeoutMs, const volatile bool &stopFlag);

        bool        cycleTransportAndServer(
                        RcfServer &server,
                        int timeoutMs,
                        const volatile bool &stopFlag);

        UdpServerTransport & enableSharedAddressBinding();

        // I_Service implementation
    private:
        void        onServiceAdded(RcfServer &server);
        void        onServiceRemoved(RcfServer &server);
        void        onServerOpen(RcfServer &server);
        void        onServerClose(RcfServer &server);
        void        onServerStart(RcfServer &server);
        void        onServerStop(RcfServer &server);

    private:

        typedef UdpSessionState SessionState;
        typedef UdpSessionStatePtr SessionStatePtr;

        I_SessionManager *  mpSessionManager;
        int                 mPort;
        int                 mFd;
        volatile bool       mStopFlag;
        unsigned int        mPollingDelayMs;
        std::string         mMulticastIp;
        bool                mEnableSharedAddressBinding;

        void        postWrite(
                        const SessionStatePtr &sessionStatePtr,
                        const std::vector<ByteBuffer> &byteBuffers);

        void        postRead(
                        const SessionStatePtr &sessionStatePtr);        

        friend class UdpProactor;

    }; // class UdpServerTransport

} // namespace RCF

#endif // ! INCLUDE_RCF_UDPSERVERTRANSPORT_HPP
