
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/BsdClientTransport.hpp>

#include <RCF/TimedBsdSockets.hpp>

namespace RCF {

    BsdClientTransport::BsdClientTransport() :
        mFd(-1)
    {}

    BsdClientTransport::BsdClientTransport(int fd) :
        mFd(fd)
    {}

    BsdClientTransport::BsdClientTransport(const BsdClientTransport & rhs) :
        ConnectionOrientedClientTransport(rhs),
        mFd(-1)
    {}

    BsdClientTransport::~BsdClientTransport()
    {
        RCF_DTOR_BEGIN
            close();
        RCF_DTOR_END
    }

#ifdef BOOST_WINDOWS

    // return -2 for timeout, -1 for error, 0 for ready
    int pollSocketWithProgressMwfmo(
        const ClientProgressPtr &clientProgressPtr,
        ClientProgress::Activity activity,
        unsigned int endTimeMs,
        int fd,
        int &err,
        bool bRead)
    {
        bool bTimer = clientProgressPtr->mTriggerMask & ClientProgress::Timer ?
            true :
            false;

        bool bUiMessage = clientProgressPtr->mTriggerMask & ClientProgress::UiMessage ?
            true :
            false;

        unsigned int pollingIntervalMs = bTimer ?
            clientProgressPtr->mTimerIntervalMs :
            -1;

        int uiMessageFilter = clientProgressPtr->mUiMessageFilter;
        HANDLE readEvent = WSACreateEvent();
        using namespace boost::multi_index::detail;
        scope_guard WSACloseEventGuard = make_guard(WSACloseEvent, readEvent);
        RCF_UNUSED_VARIABLE(WSACloseEventGuard);
        int nRet = WSAEventSelect(fd, readEvent, bRead ? FD_READ : FD_WRITE);
        RCF_ASSERT(nRet == 0)(nRet);
        HANDLE handles[] = { readEvent };
        while (true)
        {
            unsigned int timeoutMs = generateTimeoutMs(endTimeMs);
            timeoutMs = RCF_MIN(timeoutMs, pollingIntervalMs);

            DWORD dwRet = bUiMessage ?
                MsgWaitForMultipleObjects(1, handles, 0, timeoutMs, uiMessageFilter) :
                WaitForMultipleObjects(1, handles, 0, timeoutMs);

            if (dwRet == WAIT_TIMEOUT)
            {
                if (generateTimeoutMs(endTimeMs) == 0)
                {
                    err = 0;
                    return -2;
                }
                else if (bTimer)
                {
                    ClientProgress::Action action = ClientProgress::Continue;

                    clientProgressPtr->mProgressCallback(
                        0,
                        0,
                        ClientProgress::Timer,
                        activity,
                        action);

                    RCF_VERIFY(
                        action != ClientProgress::Cancel,
                        Exception(RcfError_ClientCancel))
                        (endTimeMs)(getCurrentTimeMs())(pollingIntervalMs);
                }
            }
            else if (dwRet == WAIT_OBJECT_0)
            {
                // read event signalled
                return 0;
            }
            else if (bUiMessage && dwRet == WAIT_OBJECT_0 + 1)
            {
                ClientProgress::Action action = ClientProgress::Continue;

                clientProgressPtr->mProgressCallback(
                    0,
                    0,
                    ClientProgress::UiMessage,
                    activity,
                    action);

                RCF_VERIFY(
                    action != ClientProgress::Cancel,
                    Exception(RcfError_ClientCancel))
                    (endTimeMs)(getCurrentTimeMs())(pollingIntervalMs);

                // a sample message filter

                //MSG msg = {0};
                //while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
                //{
                //    if (msg.message == WM_QUIT)
                //    {
                //
                //    }
                //    else if (msg.message == WM_PAINT)
                //    {
                //        TranslateMessage(&msg);
                //        DispatchMessage(&msg);
                //    }
                //}

            }

        }
    }

#endif

    PollingFunctor::PollingFunctor(
        const ClientProgressPtr &clientProgressPtr,
        ClientProgress::Activity activity,
        unsigned int endTimeMs) :
            mClientProgressPtr(clientProgressPtr),
            mActivity(activity),
            mEndTimeMs(endTimeMs)
    {}

    int PollingFunctor::operator()(int fd, int &err, bool bRead)
    {

#ifdef BOOST_WINDOWS

        if (
            mClientProgressPtr.get() &&
            mClientProgressPtr->mTriggerMask & ClientProgress::UiMessage)
        {
            return pollSocketWithProgressMwfmo(
                mClientProgressPtr,
                mActivity,
                mEndTimeMs,
                fd,
                err,
                bRead);
        }
        else

#endif

        if (mClientProgressPtr)
        {
            return pollSocketWithProgress(
                mClientProgressPtr,
                mActivity,
                mEndTimeMs,
                fd,
                err,
                bRead);
        }
        else
        {
            return pollSocket(
                mEndTimeMs,
                fd,
                err,
                bRead);
        }
    }

    std::size_t BsdClientTransport::implRead(
        const ByteBuffer &byteBuffer,
        std::size_t bytesRequested)
    {
        std::size_t bytesToRead = RCF_MIN(bytesRequested, byteBuffer.getLength());

        PollingFunctor pollingFunctor(
            mClientProgressPtr,
            ClientProgress::Receive,
            mEndTimeMs);

        RCF3_TRACE("calling recv()")(byteBuffer.getLength())(bytesToRead);

        int err = 0;
        int ret = RCF::timedRecv(
            pollingFunctor,
            err,
            mFd,
            byteBuffer,
            bytesToRead,
            0);

        switch (ret)
        {
        case -2:

            RCF_THROW(
                Exception(RcfError_ClientReadTimeout))
                (bytesToRead);

            break;

        case -1:
            
            RCF_THROW(
                Exception(
                    RcfError_ClientReadFail,
                    err,
                    RcfSubsystem_Os))
                (bytesToRead)(err);

            break;

        case  0:
            
            RCF_THROW(
                Exception(RcfError_PeerDisconnect))
                (bytesToRead);

            break;

        default:
            
            RCF_ASSERT(
                0 < ret && ret <= static_cast<int>(bytesRequested))
                (ret)(bytesRequested);
        }

        return ret;
    }

    std::size_t BsdClientTransport::implWrite(
        const std::vector<ByteBuffer> &byteBuffers)
    {
        PollingFunctor pollingFunctor(
            mClientProgressPtr,
            ClientProgress::Send,
            mEndTimeMs);

        int err = 0;

        RCF3_TRACE("calling send()")(lengthByteBuffers(byteBuffers));

        int ret = RCF::timedSend(
            pollingFunctor,
            err,
            mFd,
            byteBuffers,
            getMaxSendSize(),
            0);

        switch (ret)
        {
        case -2:

            RCF_THROW(
                Exception(RcfError_ClientWriteTimeout));

            break;

        case -1:
            
            RCF_THROW(
                Exception(
                    RcfError_ClientWriteFail,
                    err,
                    RcfSubsystem_Os))
                (err);

            break;

        case 0:
            
            RCF_THROW(
                Exception(RcfError_PeerDisconnect));
            
            break;

        default:
            
            RCF_ASSERT(
                0 < ret && ret <= static_cast<int>(lengthByteBuffers(byteBuffers)))
                (ret)(lengthByteBuffers(byteBuffers));            
        }

        return ret;
    }

    bool BsdClientTransport::isConnected()
    {
        return isFdConnected(mFd);
    }

    int BsdClientTransport::releaseFd()
    {
        // We've got problems if there is a close functor involved.
        RCF_ASSERT( !mNotifyCloseFunctor );

        int myFd = mFd;
        mFd = -1;
        return myFd;
    }

    int BsdClientTransport::getFd() const
    {
        return mFd;
    }

    int    BsdClientTransport::getNativeHandle() const
    {
        return mFd;
    }
    
} // namespace RCF
