
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/ConnectionOrientedClientTransport.hpp>

namespace RCF {

    class ClientFilterProxy : public Filter, boost::noncopyable
    {
    public:
        ClientFilterProxy(ConnectionOrientedClientTransport &transport, Filter &filter, bool top) :
            mTransport(transport),
            mFilter(filter),
            mTop(top)
        {}

    private:

        void read(const ByteBuffer &byteBuffer, std::size_t bytesRequested)
        {
            mTop ?
                mFilter.read(byteBuffer, bytesRequested) :
                mTransport.read(byteBuffer, bytesRequested);
        }

        void write(const std::vector<ByteBuffer> &byteBuffers)
        {
            mTop ?
                mFilter.write(byteBuffers) :
                mTransport.write(byteBuffers);
        }

        void onReadCompleted(const ByteBuffer &byteBuffer, int error)
        {
            mTop ?
                mTransport.onReadCompleted(byteBuffer, error) :
                mFilter.onReadCompleted(byteBuffer, error);
        }

        void onWriteCompleted(std::size_t bytesTransferred, int error)
        {
            mTop ?
                mTransport.onWriteCompleted(bytesTransferred, error) :
                mFilter.onWriteCompleted(bytesTransferred, error);
        }

        const FilterDescription &getFilterDescription() const
        {
            RCF_ASSERT(0);
            return * (const FilterDescription *) NULL;
        }

        ConnectionOrientedClientTransport &mTransport;
        Filter &mFilter;
        bool mTop;
    };

    ConnectionOrientedClientTransport::ConnectionOrientedClientTransport() :
        mMaxSendSize(1024*1024*10),
        mError(RCF_DEFAULT_INIT),
        mBytesTransferred(RCF_DEFAULT_INIT),
        mBytesSent(RCF_DEFAULT_INIT),
        mBytesRead(RCF_DEFAULT_INIT),
        mBytesTotal(RCF_DEFAULT_INIT),
        mEndTimeMs(RCF_DEFAULT_INIT),
        mClosed(RCF_DEFAULT_INIT)
    {
        setTransportFilters( std::vector<FilterPtr>() );
    }

    ConnectionOrientedClientTransport::ConnectionOrientedClientTransport(const ConnectionOrientedClientTransport &rhs) :
        I_ClientTransport(rhs),        
        mMaxSendSize(1024*1024*10),
        mError(RCF_DEFAULT_INIT),
        mBytesTransferred(RCF_DEFAULT_INIT),
        mBytesSent(RCF_DEFAULT_INIT),
        mBytesRead(RCF_DEFAULT_INIT),
        mBytesTotal(RCF_DEFAULT_INIT),
        mEndTimeMs(RCF_DEFAULT_INIT),
        mClosed(RCF_DEFAULT_INIT)
    {
        setTransportFilters( std::vector<FilterPtr>() );
    }

    ConnectionOrientedClientTransport::~ConnectionOrientedClientTransport()
    {
        RCF_DTOR_BEGIN
            close();
        RCF_DTOR_END
    }

    void ConnectionOrientedClientTransport::connect(unsigned int timeoutMs)
    {
        if (!isConnected())
        {
            mClosed = false;

            implConnect(timeoutMs);

            if (
                mClientProgressPtr.get() &&
                (mClientProgressPtr->mTriggerMask & ClientProgress::Event))
            {
                ClientProgress::Action action = ClientProgress::Continue;

                mClientProgressPtr->mProgressCallback(
                    0,
                    0,
                    ClientProgress::Event,
                    ClientProgress::Connect,
                    action);

                RCF_VERIFY(
                    action != ClientProgress::Cancel,
                    Exception(RcfError_ClientCancel));
            }
        }
    }

    void ConnectionOrientedClientTransport::disconnect(unsigned int timeoutMs)
    {
        RCF_UNUSED_VARIABLE(timeoutMs);

        // close the connection
        close();
    }

    void ConnectionOrientedClientTransport::read(
        const ByteBuffer &byteBuffer_, 
        std::size_t bytesRequested)
    {
        ByteBuffer byteBuffer = byteBuffer_;
        if (byteBuffer.getLength() == 0)
        {
            if (mReadBuffer2Ptr.get() == NULL || !mReadBuffer2Ptr.unique())
            {
                mReadBuffer2Ptr.reset( new std::vector<char>() );
            }
            mReadBuffer2Ptr->resize(bytesRequested);
            byteBuffer = ByteBuffer(mReadBuffer2Ptr);
        }

        std::size_t bytesRead = implRead(byteBuffer, bytesRequested);

        ByteBuffer b(byteBuffer.release(), 0, bytesRead);
        mTransportFilters.empty() ?
            onReadCompleted(b, 0) :
            mTransportFilters.back()->onReadCompleted(b, 0);
    }

    void ConnectionOrientedClientTransport::write(
        const std::vector<ByteBuffer> &byteBuffers)
    {
        std::size_t bytesWritten = implWrite(byteBuffers);

        mTransportFilters.empty() ?
            onWriteCompleted(bytesWritten, 0) :
            mTransportFilters.back()->onWriteCompleted(bytesWritten, 0);

    }

    // helper to disambiguate std::vector<ByteBuffer>::resize()
    typedef void (std::vector<ByteBuffer>::*PfnResize)(std::vector<ByteBuffer>::size_type);

    void clearByteBuffers( std::vector<ByteBuffer> &byteBuffers)
    {
        byteBuffers.resize(0);
    }

    int ConnectionOrientedClientTransport::send(
        const std::vector<ByteBuffer> &data,
        unsigned int totalTimeoutMs)
    {

        RCF3_TRACE("")(lengthByteBuffers(data))(totalTimeoutMs);

        unsigned int startTimeMs = getCurrentTimeMs();
        mEndTimeMs = startTimeMs + totalTimeoutMs;

        RCF_VERIFY(
            isConnected(),
            RCF::Exception(RcfError_NotConnected));

        mByteBuffers.resize(0);

        using namespace boost::multi_index::detail;
        scope_guard resizeGuard =
            make_guard(clearByteBuffers, boost::ref(mByteBuffers));
        RCF_UNUSED_VARIABLE(resizeGuard);

        std::copy(data.begin(), data.end(), std::back_inserter(mByteBuffers));

        ByteBuffer &byteBuffer = const_cast<ByteBuffer &>(mByteBuffers.front());
        unsigned int messageLength = static_cast<unsigned int>(lengthByteBuffers(data));
        BOOST_STATIC_ASSERT(sizeof(unsigned int) == 4);
        RCF_ASSERT(byteBuffer.getLeftMargin() >= 4)(byteBuffer.getLeftMargin());
        byteBuffer.expandIntoLeftMargin(4);
        memcpy(byteBuffer.getPtr(), &messageLength, 4);
        RCF::machineToNetworkOrder(byteBuffer.getPtr(), 4, 1);

        std::size_t ret = timedSend(mByteBuffers);
        RCF_ASSERT(ret == lengthByteBuffers(mByteBuffers))(ret)(lengthByteBuffers(mByteBuffers));

        return 1;
    }

    std::size_t ConnectionOrientedClientTransport::timedSend(const std::vector<ByteBuffer> &data)
    {
        std::size_t bytesRequested = lengthByteBuffers(data);
        std::size_t bytesToWrite = bytesRequested;
        std::size_t bytesWritten = 0;

        using namespace boost::multi_index::detail;
        scope_guard resizeGuard =
            make_guard(clearByteBuffers, boost::ref(mSlicedByteBuffers));
        RCF_UNUSED_VARIABLE(resizeGuard);

        while (true)
        {
            sliceByteBuffers(mSlicedByteBuffers, data, bytesWritten);

            mTransportFilters.empty() ?
                write(mSlicedByteBuffers):
                mTransportFilters.front()->write(mSlicedByteBuffers);

            RCF_ASSERT(
                0 < mBytesTransferred &&
                mBytesTransferred <= lengthByteBuffers(mSlicedByteBuffers))
                (mBytesTransferred)(lengthByteBuffers(mSlicedByteBuffers));

            bytesToWrite -= mBytesTransferred;
            bytesWritten += mBytesTransferred;

            if (
                mClientProgressPtr.get() &&
                (mClientProgressPtr->mTriggerMask & ClientProgress::Event))
            {
                ClientProgress::Action action = ClientProgress::Continue;

                mClientProgressPtr->mProgressCallback(
                    bytesWritten, bytesRequested,
                    ClientProgress::Event,
                    ClientProgress::Send,
                    action);

                RCF_VERIFY(
                    action != ClientProgress::Cancel,
                    Exception(RcfError_ClientCancel))
                    (mBytesSent)(mBytesTotal);
            }

            if (bytesToWrite == 0)
            {
                return bytesWritten;
            }
        }
    }

    // return bufferLen
    std::size_t ConnectionOrientedClientTransport::timedReceive(
        ByteBuffer &byteBuffer,
        std::size_t bytesRequested)
    {
        std::size_t bytesToRead = bytesRequested;
        std::size_t bytesRead = 0;
        while (true)
        {
            ByteBuffer buffer(byteBuffer, bytesRead, bytesToRead);

            mTransportFilters.empty() ?
                read(buffer, bytesToRead):
                mTransportFilters.front()->read(buffer, bytesToRead);

            RCF_ASSERT(
                0 < mBytesTransferred && mBytesTransferred <= bytesToRead)
                (mBytesTransferred)(bytesRead);

            bytesToRead -= mBytesTransferred;
            bytesRead += mBytesTransferred;

            if (
                mClientProgressPtr.get() &&
                (mClientProgressPtr->mTriggerMask & ClientProgress::Event))
            {
                ClientProgress::Action action = ClientProgress::Continue;

                mClientProgressPtr->mProgressCallback(
                    bytesRead,
                    bytesRequested,
                    ClientProgress::Event,
                    ClientProgress::Receive,
                    action);

                RCF_VERIFY(
                    action != ClientProgress::Cancel,
                    Exception(RcfError_ClientCancel))
                    (mBytesRead)(mBytesTotal);
            }

            if (bytesToRead == 0)
            {
                return bytesRead;
            }

        }
    }

    int ConnectionOrientedClientTransport::receive(
        ByteBuffer &byteBuffer,
        unsigned int timeoutMs)
    {
        mEndTimeMs = getCurrentTimeMs() + timeoutMs;

        unsigned int length = 0;
        BOOST_STATIC_ASSERT(sizeof(length) == 4);
        ByteBuffer buffer( (char*) &length, 4);
        std::size_t ret = timedReceive(buffer, 4);

        RCF_ASSERT(ret == 4)(ret);
        networkToMachineOrder(&length, sizeof(length), 1);
        RCF_VERIFY(
            0 <= length && length <= getMaxMessageLength(),
            Exception(RcfError_ClientMessageLength))
            (length)(getMaxMessageLength());

        // read the rest of the message
        if (mReadBufferPtr.get() == NULL || !mReadBufferPtr.unique())
        {
            mReadBufferPtr.reset( new std::vector<char>() );
        }
        mReadBufferPtr->resize(length);

        buffer = ByteBuffer(
            &(*mReadBufferPtr)[0],
            mReadBufferPtr->size(),
            mReadBufferPtr);

        ret = timedReceive(buffer, length);

        RCF_ASSERT(ret == length);
        byteBuffer = buffer;
        RCF3_TRACE("")(length);
        return 1;
    }

    void ConnectionOrientedClientTransport::setCloseFunctor(const CloseFunctor &closeFunctor)
    {
        mCloseFunctor = closeFunctor;
    }

    void ConnectionOrientedClientTransport::setNotifyCloseFunctor(const NotifyCloseFunctor &notifyCloseFunctor)
    {
        mNotifyCloseFunctor = notifyCloseFunctor;
    }

    void ConnectionOrientedClientTransport::close()
    {
        if (!mClosed)
        {

            // First notify anyone who is interested.
            if (mNotifyCloseFunctor)
            {
                mNotifyCloseFunctor();
                mNotifyCloseFunctor = NotifyCloseFunctor();
            }

            // And then figure out how to close the socket.
            if (mCloseFunctor)
            {
                // Actually just a workaround for TcpAsioServerTransport.
                mCloseFunctor();
                mCloseFunctor = CloseFunctor();
            }
            else
            {
                implClose();            
            }

            mClosed = true;
        }
    }

    void ConnectionOrientedClientTransport::setTransportFilters(
        const std::vector<FilterPtr> &filters)
    {
        mTransportFilters.clear();
        if (!filters.empty())
        {
            mTransportFilters.push_back(
                FilterPtr(new ClientFilterProxy(*this, *filters.front(), true)));

            std::copy(
                filters.begin(),
                filters.end(),
                std::back_inserter(mTransportFilters));

            mTransportFilters.push_back(
                FilterPtr(new ClientFilterProxy(*this, *filters.back(), false)));

            RCF::connectFilters(mTransportFilters);

        }
    }

    void ConnectionOrientedClientTransport::onReadCompleted(
        const ByteBuffer &byteBuffer,
        int error)
    {
        if (error)
        {
            RCF_THROW(Exception(RcfError_Filter));
        }
        else
        {
            mBytesTransferred = byteBuffer.getLength() ;
            mError = error;
        }
    }
    void ConnectionOrientedClientTransport::getTransportFilters(
        std::vector<FilterPtr> &filters)
    {
        // TODO: keep the adapter filters out of mTransportFilters?
        if (mTransportFilters.empty())
        {
            filters.resize(0);
        }
        else
        {
            RCF_ASSERT(mTransportFilters.size() >= 3)(mTransportFilters.size());

            //filters.assign(
            //    ++mTransportFilters.begin(),
            //    --mTransportFilters.end());

            // for borland
            std::vector<FilterPtr>::const_iterator iter0(mTransportFilters.begin());
            std::vector<FilterPtr>::const_iterator iter1(mTransportFilters.end());
            ++iter0;
            --iter1;
            filters.assign(iter0, iter1);

        }
    }

    void ConnectionOrientedClientTransport::onWriteCompleted(
        std::size_t bytesTransferred,
        int error)
    {
        if (error)
        {
            // TODO: this should never happen, use an assertion
            RCF_THROW(Exception(RcfError_Filter));
        }
        else
        {
            mBytesTransferred = bytesTransferred;
            mError = error;
        }
    }

    void ConnectionOrientedClientTransport::setMaxSendSize(std::size_t maxSendSize)
    {
        mMaxSendSize = maxSendSize;
    }

    std::size_t ConnectionOrientedClientTransport::getMaxSendSize()
    {
        return mMaxSendSize;
    }

} // namespace RCF
