
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/SubscriptionService.hpp>

#include <boost/bind.hpp>

#include <typeinfo>

#include <RCF/ClientTransport.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/RcfSession.hpp>
#include <RCF/ServerInterfaces.hpp>

namespace RCF {

    bool Subscription::isConnected()
    {
        return
            mClientTransportAutoPtr.get() &&
            mClientTransportAutoPtr->isConnected();
    }

    void Subscription::close()
    {
        {
            RcfSessionPtr rcfSessionPtr(mRcfSessionWeakPtr.lock());
            if (rcfSessionPtr)
            {
                rcfSessionPtr->setOnDestroyCallback(RcfSession::OnDestroyCallback());
            }
        }
        mRcfSessionWeakPtr.reset();
        mClientTransportAutoPtr.reset();
    }

    SubscriptionService::SubscriptionService() :
        mSubscriptionsMutex(WriterPriority)
    {}

    void vc6_boost_bind_helper(SubscriptionService::OnDisconnect onDisconnect, RcfSession &)
    {
        onDisconnect();
    }

    SubscriptionPtr SubscriptionService::beginSubscribeNamed(
        SubscriptionId whichSubscription,
        RcfClientPtr rcfClientPtr,
        ClientStub & clientStub,
        OnDisconnect onDisconnect,
        const std::string &publisherName)
    {
        I_ServerTransportEx &serverTransportEx =
            dynamic_cast<I_ServerTransportEx &>(*mServerTransportPtr);

        RcfClient<I_RequestSubscription> client(clientStub);
        client.getClientStub().setTransport(clientStub.releaseTransport());
        
        bool ok = (RcfError_Ok == client.requestSubscription(Twoway, publisherName));
        if (ok)
        {
            SessionPtr sessionPtr =
                serverTransportEx.createServerSession(
                    client.getClientStub().releaseTransport(),
                    StubEntryPtr(new StubEntry(rcfClientPtr)));

            RCF_ASSERT( sessionPtr );

            RcfSessionPtr rcfSessionPtr = 
                boost::static_pointer_cast<RcfSession>(sessionPtr);

            rcfSessionPtr->setUserData(client.getClientStub().getUserData());

            if (onDisconnect)
            {

#if defined(_MSC_VER) && _MSC_VER < 1310

                rcfSessionPtr->setOnDestroyCallback(
                    boost::bind(vc6_boost_bind_helper, onDisconnect, _1));

#else

                rcfSessionPtr->setOnDestroyCallback(
                    boost::bind(onDisconnect));

#endif

            }          
            ClientTransportAutoPtr apClientTransport(
                serverTransportEx.createClientTransport(sessionPtr) );

            SubscriptionPtr subscriptionPtr(
                new Subscription(apClientTransport, rcfSessionPtr));

            WriteLock lock(mSubscriptionsMutex);
            mSubscriptions[ whichSubscription ] = subscriptionPtr;                
            return subscriptionPtr;                
        }
        return SubscriptionPtr();
    }

    SubscriptionPtr SubscriptionService::beginSubscribeNamed(
        SubscriptionId whichSubscription,
        RcfClientPtr rcfClientPtr,
        const I_Endpoint &publisherEndpoint,
        OnDisconnect onDisconnect,
        const std::string &publisherName)
    {
        ClientStub clientStub("");
        clientStub.setEndpoint(publisherEndpoint);
        clientStub.createTransport();

        return
            beginSubscribeNamed(
                whichSubscription,
                rcfClientPtr,
                clientStub,
                onDisconnect,
                publisherName);
    }

    SubscriptionPtr SubscriptionService::beginSubscribeNamed(
        SubscriptionId whichSubscription,
        RcfClientPtr rcfClientPtr,
        ClientTransportAutoPtr clientTransportAutoPtr,
        OnDisconnect onDisconnect,
        const std::string &publisherName)
    {
        ClientStub clientStub("");
        clientStub.setTransport(clientTransportAutoPtr);

        return
            beginSubscribeNamed(
                whichSubscription,
                rcfClientPtr,
                clientStub,
                onDisconnect,
                publisherName);

    }

    bool SubscriptionService::endSubscribeNamed(SubscriptionId whichSubscription)
    {
        SubscriptionPtr subscriptionPtr;
        {
            WriteLock lock(mSubscriptionsMutex);
            subscriptionPtr = mSubscriptions[ whichSubscription ];
            mSubscriptions.erase(whichSubscription);
        }
        if (subscriptionPtr)
        {
            RcfSessionPtr rcfSessionPtr =
                subscriptionPtr->mRcfSessionWeakPtr.lock();

            if (rcfSessionPtr)
            {
                // block server threads from executing remote calls
                Lock lock(rcfSessionPtr->mStopCallInProgressMutex);
                rcfSessionPtr->mStopCallInProgress = true;

                // remove subscription binding
                rcfSessionPtr->setDefaultStubEntryPtr(StubEntryPtr());

                // clear the destroy callback
                // TODO: how do we know that we're not clearing someone else's callback?
                rcfSessionPtr->setOnDestroyCallback(RcfSession::OnDestroyCallback());
            }
        }
        return true;
    }

    SubscriptionPtr SubscriptionService::getSubscriptionPtr(
        SubscriptionId whichSubscription)
    {
        ReadLock lock(mSubscriptionsMutex);
        Subscriptions::iterator iter = mSubscriptions.find(whichSubscription);
        return iter != mSubscriptions.end() ? 
            iter->second : 
            SubscriptionPtr();
    }

    void SubscriptionService::onServiceAdded(RcfServer &server)
    {
        mServerTransportPtr = server.getServerTransportPtr();
    }

    void SubscriptionService::onServiceRemoved(RcfServer &)
    {}

    void SubscriptionService::onServerStop(RcfServer &server)
    {
        RCF_UNUSED_VARIABLE(server);
        WriteLock writeLock(mSubscriptionsMutex);

        for (Subscriptions::iterator iter = mSubscriptions.begin();
            iter != mSubscriptions.end();
            ++iter)
        {
            SubscriptionPtr subscriptionPtr = iter->second;
            subscriptionPtr->close();
        }

        mSubscriptions.clear();
    }

    void SubscriptionService::onServerClose(RcfServer &)
    {}

    Subscription::Subscription(
        ClientTransportAutoPtr clientTransportAutoPtr,
        RcfSessionWeakPtr rcfSessionWeakPtr) :
            mClientTransportAutoPtr(clientTransportAutoPtr),
            mRcfSessionWeakPtr(rcfSessionWeakPtr)
    {}
   
} // namespace RCF
