
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/TcpClientTransport.hpp>

#include <RCF/TimedBsdSockets.hpp>
#include <RCF/TcpEndpoint.hpp>

namespace RCF {

    TcpClientTransport::TcpClientTransport(const TcpClientTransport &rhs) : 
        BsdClientTransport(rhs),
        mRemoteAddr(rhs.mRemoteAddr),
        mIp(rhs.mIp),
        mPort(rhs.mPort)
    {}

    TcpClientTransport::TcpClientTransport(const std::string &ip, int port) :
        BsdClientTransport(),
        mRemoteAddr(),
        mIp(ip),
        mPort(port)
    {
        memset(&mRemoteAddr, 0, sizeof(mRemoteAddr));
    }

    TcpClientTransport::TcpClientTransport(const sockaddr_in &remoteAddr) :
        BsdClientTransport(),
        mRemoteAddr(remoteAddr),
        mIp(),
        mPort(RCF_DEFAULT_INIT)
    {}

    TcpClientTransport::TcpClientTransport(int fd) :
        BsdClientTransport(fd),
        mRemoteAddr(),
        mIp(),
        mPort(RCF_DEFAULT_INIT)
    {
        memset(&mRemoteAddr, 0, sizeof(mRemoteAddr));
    }

    int TcpClientTransport::getSocket()
    {
        return mFd;
    }

    TcpClientTransport::~TcpClientTransport()
    {
        RCF_DTOR_BEGIN
            close();
        RCF_DTOR_END
    }

    std::auto_ptr<I_ClientTransport> TcpClientTransport::clone() const
    {
        return ClientTransportAutoPtr( new TcpClientTransport(*this) );
    }

    void TcpClientTransport::implConnect(unsigned int timeoutMs)
    {
        // TODO: replace throw with return, where possible
        
        // close the current connection
        implClose();

        mFd = static_cast<int>( ::socket(PF_INET, SOCK_STREAM, IPPROTO_TCP) );
        int err = Platform::OS::BsdSockets::GetLastError();
        RCF_VERIFY(
            mFd != -1,
            Exception(
                RcfError_Socket, err, RcfSubsystem_Os, "socket() failed"));
        Platform::OS::BsdSockets::setblocking(mFd, false);

        if (mRemoteAddr.sin_addr.s_addr == 0)
        {
            unsigned long ul_addr = ::inet_addr(mIp.c_str());
            if (ul_addr == INADDR_NONE)
            {
                hostent *hostDesc = ::gethostbyname( mIp.c_str() );
                if (hostDesc)
                {
                    char *szIp = ::inet_ntoa( * (in_addr*) hostDesc->h_addr_list[0]);
                    ul_addr = ::inet_addr(szIp);
                }
            }

            memset(&mRemoteAddr, 0, sizeof(mRemoteAddr));
            mRemoteAddr.sin_family = AF_INET;
            mRemoteAddr.sin_addr.s_addr = ul_addr;
            // the :: seems to screw up gcc (!?!)
            //remoteAddr.sin_port = ::htons(mPort);
            mRemoteAddr.sin_port = htons( static_cast<u_short>(mPort) );
        }

        unsigned int startTimeMs = getCurrentTimeMs();
        mEndTimeMs = startTimeMs + timeoutMs;

        PollingFunctor pollingFunctor(
            mClientProgressPtr,
            ClientProgress::Connect,
            mEndTimeMs);

        err = 0;

        int ret = timedConnect(
            pollingFunctor,
            err,
            mFd,
            (sockaddr*) &mRemoteAddr,
            sizeof(mRemoteAddr));

        if (ret != 0)
        {
            implClose();

            int rcfErr = (err == 0) ?
                RcfError_ClientConnectTimeout :
                RcfError_ClientConnectFail;

            RCF_THROW(
                Exception(rcfErr, err, RcfSubsystem_Os))
                (ret)(err)(timeoutMs)(mIp)(mPort);
        }

    }

    void TcpClientTransport::implClose()
    {
        if (mFd != -1)
        {
            int ret = Platform::OS::BsdSockets::closesocket(mFd);
            int err = Platform::OS::BsdSockets::GetLastError();

            RCF_VERIFY(
                ret == 0,
                Exception(
                    RcfError_Socket,
                    err,
                    RcfSubsystem_Os,
                    "closesocket() failed"))
                (mFd);
        }

        mFd = -1;
    }

    EndpointPtr TcpClientTransport::getEndpointPtr() const
    {
        return EndpointPtr( new TcpEndpoint(mIp, mPort) );
    }

    void TcpClientTransport::setRemoteAddr(const sockaddr_in &remoteAddr)
    {
        mRemoteAddr = remoteAddr;
    }

    const sockaddr_in &TcpClientTransport::getRemoteAddr() const
    {
        return mRemoteAddr;
    }

} // namespace RCF
