
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 1.0
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/Win32NamedPipeServerTransport.hpp>

#include <RCF/ByteOrdering.hpp>
#include <RCF/CurrentSession.hpp>
#include <RCF/MethodInvocation.hpp>
#include <RCF/RcfServer.hpp>

namespace RCF {

    Win32NamedPipeOverlapped::Win32NamedPipeOverlapped(
        HANDLE hPipe, 
        Win32NamedPipeServerTransport & serverTransport) :
            mhPipe(hPipe),
            mState(Connecting),
            mReadBufferPos(),
            mWriteBufferPos(),
            mServerTransport(serverTransport)
    {
        memset(static_cast<OVERLAPPED *>(this), 0, sizeof(OVERLAPPED));
    }

    Win32NamedPipeOverlapped::~Win32NamedPipeOverlapped()
    {
        // No possibility of exceptions in this dtor.
        //RCF_DTOR_BEGIN
            FlushFileBuffers(mhPipe); 
            DisconnectNamedPipe(mhPipe); 
            CloseHandle(mhPipe); 
        //RCF_DTOR_END
    }

    void Win32NamedPipeServerTransport::setSecurityAttributes(
        LPSECURITY_ATTRIBUTES pSec)
    {
        mpSec = pSec;
    }

    void Win32NamedPipeOverlapped::reconnect()
    {
        mSessionPtr.reset();

        BOOL ok = DisconnectNamedPipe(mhPipe);
        DWORD dwErr = GetLastError();
        RCF_VERIFY(ok, Exception(RcfError_Pipe, dwErr));

        mState = Connecting;
        ok = ConnectNamedPipe(mhPipe, this);
        dwErr = GetLastError();
        RCF_VERIFY(!ok && dwErr == WSA_IO_PENDING, Exception(RcfError_Pipe, dwErr));
    }

    void Win32NamedPipeOverlapped::issueRead()
    {
        mState = Reading;

        DWORD dwBytesRead = 0;
        BOOL ok = ReadFile(
            mhPipe, 
            &(*mReadBufferPtr)[mReadBufferPos], 
            static_cast<DWORD>(mReadBufferPtr->size() - mReadBufferPos),
            &dwBytesRead, 
            this);

        DWORD dwErr = GetLastError();

        if (dwBytesRead)
        {
        }
        else
        {
            RCF_ASSERT(!ok);
            if (
                dwErr == ERROR_IO_PENDING || 
                dwErr == WSA_IO_PENDING || 
                dwErr == ERROR_MORE_DATA)
            {
                // Nothing wrong
            }
            else
            {
                reconnect();
            }
        }
    }

    void Win32NamedPipeOverlapped::issueWrite()
    {
        mState = Writing;

        ByteBuffer byteBuffer = sliceByteBuffer(mWriteBuffers, mWriteBufferPos);

        DWORD bytesWritten = 0;
        BOOL ok = WriteFile(
            mhPipe,
            byteBuffer.getPtr(),
            static_cast<DWORD>(byteBuffer.getLength()),
            &bytesWritten,
            this);

        DWORD dwErr = GetLastError();

        if (bytesWritten)
        {
        }
        else
        {
            RCF_ASSERT(!ok);
            if (
                dwErr == ERROR_IO_PENDING || 
                dwErr == WSA_IO_PENDING || 
                dwErr == ERROR_MORE_DATA)
            {
                // Nothing wrong
            }
            else
            {
                reconnect();
            }
        }
    }

    void Win32NamedPipeOverlapped::onCompletion(
        std::size_t bytesTransferred)
    {
        switch (mState)
        {
        case Connecting:
            
            mSessionPtr = mServerTransport.mpSessionManager->createSession();
            mSessionPtr->setProactorPtr( 
                boost::shared_ptr<I_Proactor>(this, NullDeleter()));

            postRead();

            break;

        case Reading:
            if (bytesTransferred == 0)
            {
                // Client disconnected
                reconnect();
            }
            else
            {
                mReadBufferPos += bytesTransferred;
                if (mReadBufferPos < 4)
                {
                    issueRead();
                }
                else if (mReadBufferPos == 4)
                {
                    unsigned int msgLength = 0;
                    memcpy( &msgLength, &(*mReadBufferPtr)[0], 4);
                    RCF::networkToMachineOrder(&msgLength, 4, 1);

                    if (msgLength <= mServerTransport.getMaxMessageLength())
                    {
                        mReadBufferPtr->resize(4+msgLength);
                        issueRead();
                    }
                    else
                    {
                        // Do a hard close on the connection. 
                        
                        // It would be  nice to send an informative message back 
                        // to the client, as the TCP transports do, but it seems 
                        // rather difficult to do so reliably. If we write 
                        // something to the pipe, and then call FlushFileBuffers(), 
                        // we could potentially block indefinitely.

                        reconnect();

                        /*
                        ByteBuffer byteBuffer = encodeServerError(RcfError_ServerMessageLength);

                        // TODO: asynchronous return of error message failing in 
                        // release mode on vc8. For now use synchronous return.
                        bool synchronousReply = true;
                        if (synchronousReply)
                        {
                            unsigned int len = static_cast<unsigned int>(
                                byteBuffer.getLength());
                            byteBuffer.expandIntoLeftMargin(4);
                            memcpy(byteBuffer.getPtr(), &len, 4);
                            machineToNetworkOrder(byteBuffer.getPtr(), 4, 1);

                            DWORD dwWritten = 0;

                            BOOL ok = WriteFile(
                                mhPipe, 
                                byteBuffer.getPtr(), 
                                static_cast<DWORD>(byteBuffer.getLength()), 
                                &dwWritten, 
                                NULL);
                            
                            DWORD dwErr = GetLastError();
                            RCF_UNUSED_VARIABLE(dwErr);

                            if (!ok || dwWritten != byteBuffer.getLength())
                            {
                                reconnect();
                            }
                        }
                        else
                        {
                            std::vector<ByteBuffer> byteBuffers;
                            byteBuffers.push_back(byteBuffer);
                            postWrite(byteBuffers);
                        }
                        */
                    }
                }
                else if (mReadBufferPos < mReadBufferPtr->size())
                {
                    issueRead();
                }
                else if (mReadBufferPos == mReadBufferPtr->size())
                {
                    mServerTransport.mpSessionManager->onReadCompleted(mSessionPtr);
                }
                else
                {
                    RCF_ASSERT(0);
                }
            }

            break;

        case Writing:

            mWriteBufferPos += bytesTransferred;

            if (mWriteBufferPos < lengthByteBuffers(mWriteBuffers))
            {
                issueWrite();
            }
            else if (mWriteBufferPos == lengthByteBuffers(mWriteBuffers))
            {
                mWriteBuffers.resize(0);
                mServerTransport.mpSessionManager->onWriteCompleted(mSessionPtr);
            }

            break;

        default:
            RCF_ASSERT(0);
        }
    }

    // I_Proactor members

    void Win32NamedPipeOverlapped::postRead()
    {
        if (!mReadBufferPtr || !mReadBufferPtr.unique())
        {
            mReadBufferPtr.reset( new std::vector<char>());
        }

        mReadBufferPtr->resize(4);
        mReadBufferPos = 0;
        issueRead();
    }

    ByteBuffer Win32NamedPipeOverlapped::getReadByteBuffer()
    {
        return ByteBuffer( ByteBuffer(mReadBufferPtr), 4);
    }

    void Win32NamedPipeOverlapped::postWrite(
        const std::vector<ByteBuffer> &byteBuffers)
    {
        mWriteBuffers.resize(0);

        std::copy(
            byteBuffers.begin(),
            byteBuffers.end(),
            std::back_inserter(mWriteBuffers));

        int messageSize = static_cast<int>(RCF::lengthByteBuffers(byteBuffers));
        ByteBuffer &byteBuffer = mWriteBuffers.front();

        RCF_ASSERT(
            byteBuffer.getLeftMargin() >= 4)
            (byteBuffer.getLeftMargin());
        
        byteBuffer.expandIntoLeftMargin(4);
        memcpy(byteBuffer.getPtr(), &messageSize, 4);
        RCF::machineToNetworkOrder(byteBuffer.getPtr(), 4, 1);

        mWriteBufferPos = 0;
        issueWrite();
    }

    void Win32NamedPipeOverlapped::postClose()
    {
        reconnect();
    }

    I_ServerTransport &
    Win32NamedPipeOverlapped::getServerTransport()
    {
        return mServerTransport;
    }

    const I_RemoteAddress &
    Win32NamedPipeOverlapped::getRemoteAddress()
    {
        return mRemoteAddress;
    }

    void Win32NamedPipeOverlapped::setTransportFilters(
        const std::vector<FilterPtr> &filters)
    {
        RCF_UNUSED_VARIABLE(filters);

        // Not implemented
    }

    void Win32NamedPipeOverlapped::getTransportFilters(
        std::vector<FilterPtr> &)
    {
        // Not implemented
        RCF_ASSERT(0);
    }

    HANDLE    Win32NamedPipeOverlapped::getNativeHandle() const
    {
        return mhPipe;
    }

    // Win32NamedPipeServerTransport

    Win32NamedPipeServerTransport::Win32NamedPipeServerTransport(
        const tstring & pipeName,
        const tstring & machine) :
            mpSessionManager(),
            mStopFlag(RCF_DEFAULT_INIT),
            mIocp(),
            mpSec(RCF_DEFAULT_INIT)
    {
        if (pipeName.at(0) == RCF_T('\\'))
        {
            mPipeName = pipeName;
        }
        else if (machine.empty())
        {
            mPipeName = RCF_T("\\\\.\\pipe\\") + pipeName;
        }
        else
        {
            mPipeName = RCF_T("\\\\") + machine + RCF_T("\\pipe\\") + pipeName;
        }
    }

    Win32NamedPipeServerTransport::~Win32NamedPipeServerTransport()
    {
        RCF_DTOR_BEGIN
            close();
        RCF_DTOR_END
    }

    ServerTransportPtr Win32NamedPipeServerTransport::clone()
    {
        return ServerTransportPtr( 
            new Win32NamedPipeServerTransport(mPipeName));
    }

    bool Win32NamedPipeServerTransport::cycleTransportAndServer(
        RcfServer &server,
        int timeoutMs,
        const volatile bool &stopFlag)
    {
        if (!stopFlag && !mStopFlag)
        {
            cycle(timeoutMs/2, stopFlag);
            server.cycleSessions(timeoutMs/2, stopFlag);
        }
        return stopFlag || mStopFlag;
    }

    void Win32NamedPipeServerTransport::cycle(
        int timeoutMs,
        const volatile bool &stopFlag)
    {
        RCF_UNUSED_VARIABLE(stopFlag);
         
        // extract a completed io operation from the iocp
        DWORD           dwMilliseconds = timeoutMs < 0 ? INFINITE : timeoutMs;
        DWORD           dwNumBytes = 0;
        ULONG_PTR       completionKey = 0;
        OVERLAPPED *    pOverlapped = 0;

        BOOL ret = mIocp->GetStatus(
            &completionKey,
            &dwNumBytes,
            &pOverlapped,
            dwMilliseconds);

        DWORD dwErr = GetLastError();

        RCF_UNUSED_VARIABLE(ret);

        RCF_ASSERT(
            pOverlapped || (!pOverlapped && dwErr == WAIT_TIMEOUT))
            (pOverlapped)(dwErr);

        if (pOverlapped)
        {
            Win32NamedPipeOverlapped *pOverlappedPipe = static_cast<Win32NamedPipeOverlapped *>(pOverlapped);
            std::size_t bytesTransferred = dwNumBytes;
            pOverlappedPipe->onCompletion(bytesTransferred);
        }
    }

    void Win32NamedPipeServerTransport::open()
    {
        const std::size_t       MaxPipeInstances    = PIPE_UNLIMITED_INSTANCES;
        const DWORD             OutBufferSize       = 4096;
        const DWORD             InBufferSize        = 4096;
        const DWORD             DefaultTimeoutMs    = 0;

        mIocp.reset( new TcpIocp::Iocp(0));

        for (std::size_t i=0; i<MaxPipeInstances; ++i)
        {
            HANDLE hPipe = CreateNamedPipe( 
                mPipeName.c_str(),
                PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
                PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT,
                MaxPipeInstances,
                OutBufferSize,
                InBufferSize,
                DefaultTimeoutMs,
                mpSec);

            DWORD dwErr = GetLastError();

            RCF_VERIFY(hPipe != INVALID_HANDLE_VALUE, Exception(RcfError_Pipe, dwErr));

            BOOL ok = mIocp->AssociateDevice(hPipe, (ULONG_PTR) hPipe);
            dwErr = GetLastError();
            RCF_ASSERT(ok);

            Win32NamedPipeOverlapped *pOverlapped = new Win32NamedPipeOverlapped(hPipe, *this);

            mPipes.push_back(pOverlapped);

            pOverlapped->reconnect();            
        }
    }

    void Win32NamedPipeServerTransport::close()
    {
        for (std::size_t i=0; i<mPipes.size(); ++i)
        {
            delete mPipes[i];
            mPipes[i] = NULL;
        }
        mPipes.clear();
        mIocp.reset();
    }

    void Win32NamedPipeServerTransport::onServiceAdded(RcfServer &server)
    {
        mpSessionManager = &server;

        WriteLock writeLock( getTaskEntriesMutex() );
        getTaskEntries().clear();

        getTaskEntries().push_back(
            TaskEntry(
                boost::bind(
                    &Win32NamedPipeServerTransport::cycleTransportAndServer,
                    this,
                    boost::ref(server),
                    _1,
                    _2),
                StopFunctor(),
                "RCF Win32 named pipe worker"));

        mStopFlag = false;
    }

    void Win32NamedPipeServerTransport::onServiceRemoved(RcfServer &)
    {
    }

    void Win32NamedPipeServerTransport::onServerOpen(RcfServer &)
    {
        open();
    }

    void Win32NamedPipeServerTransport::onServerClose(RcfServer &)
    {
        close();
    }

    void Win32NamedPipeServerTransport::onServerStart(RcfServer &)
    {
    }

    void Win32NamedPipeServerTransport::onServerStop(RcfServer &)
    {
        mStopFlag = false;
    }

    Win32NamedPipeImpersonator::Win32NamedPipeImpersonator()
    {
        impersonate();
    }

    Win32NamedPipeImpersonator::~Win32NamedPipeImpersonator()
    {
        RCF_DTOR_BEGIN
            revertToSelf();
        RCF_DTOR_END
    }

    void Win32NamedPipeImpersonator::impersonate()
    {
        RCF::ProactorPtr proactorPtr = 
            RCF::getCurrentSessionPtr()->getProactorPtr();

        RCF::Win32NamedPipeOverlapped & pipe = 
            dynamic_cast<RCF::Win32NamedPipeOverlapped &>(
            *proactorPtr);

        BOOL ok = ImpersonateNamedPipeClient(pipe.mhPipe);
        DWORD dwErr = GetLastError();
        RCF_VERIFY(ok, RCF::Exception(RcfError_Pipe, dwErr));
    }

    void Win32NamedPipeImpersonator::revertToSelf() const
    {
        BOOL ok = RevertToSelf();
        DWORD dwErr = GetLastError();
        RCF_VERIFY(ok, RCF::Exception(RcfError_Pipe, dwErr));
    }
    
} // namespace RCF
