
#include <iostream>

#include <RCF/test/TestMinimal.hpp>

#include <RCF/Idl.hpp>
#include <RCF/IpServerTransport.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/TcpEndpoint.hpp>
#include <RCF/UdpEndpoint.hpp>

#include <RCF/test/TransportFactories.hpp>
#include <RCF/util/Platform/OS/Sleep.hpp>

RCF_BEGIN(I_X, "I_X")
RCF_METHOD_V1(void, ping, const std::string &)
RCF_END(I_X)

RCF::Mutex gMutex;
int gPings = 0;

class X
{
public:
    void ping(const std::string &s)
    {
        RCF::Lock lock(gMutex);
        std::cout << s << std::endl;
        ++gPings;
    }   
};

RCF_BEGIN(I_Broadcast, "I_Broadcast")
RCF_METHOD_V1(void, serverRunningOnPort, int)
RCF_END(I_Broadcast)

class Broadcast
{
public:
    Broadcast() : mPort(RCF_DEFAULT_INIT)
    {
    }

    void serverRunningOnPort(int port)
    {
        mPort = port;
    }

    int mPort;
};


void broadcastTask(
    int port,
    const std::string &multicastIp,
    int multicastPort,
    const RCF::RcfServer &server)
{
    RcfClient<I_Broadcast> broadcast(RCF::UdpEndpoint(multicastIp, multicastPort));
    broadcast.getClientStub().setDefaultCallingSemantics(RCF::Oneway);

    while (!server.getStopFlag())
    {
        broadcast.serverRunningOnPort(port);
        Platform::OS::SleepMs(500);
    }
}

#define GCC_3_3_1_HACK (const RCF::I_Endpoint &)

int test_main(int argc, char **argv)
{
    {
        X x;
        RCF::RcfServer server( RCF::UdpEndpoint("0.0.0.0", 0));
        server.bind( (I_X *) NULL, x);
        server.start();
        
        int udpPort = server.getIpServerTransport().getPort();

        gPings = 0;
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("127.0.0.1", udpPort)).ping(RCF::Oneway, "asdf");
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("232.5.5.5", udpPort)).ping(RCF::Oneway, "asdf");
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("232.5.5.5", udpPort+1)).ping(RCF::Oneway, "asdf");
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("232.5.5.6", udpPort)).ping(RCF::Oneway, "asdf");
        Platform::OS::SleepMs(1000);
        BOOST_CHECK(gPings == 1);

        gPings = 0;
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("255.255.255.255", udpPort)).ping(RCF::Oneway, "asdf");
        Platform::OS::SleepMs(1000);
        BOOST_CHECK(gPings >= 1);
    }

    {
        X x;
        
        RCF::RcfServer server( 
            RCF::UdpEndpoint("0.0.0.0", 0).listenOnMulticast("232.5.5.5"));

        server.bind( (I_X *) NULL, x);
        server.start();
        
        int udpPort = server.getIpServerTransport().getPort();

        gPings = 0;
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("127.0.0.1", udpPort)).ping(RCF::Oneway, "asdf");
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("232.5.5.5", udpPort)).ping(RCF::Oneway, "asdf");
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("232.5.5.5", udpPort+1)).ping(RCF::Oneway, "asdf");
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("232.5.5.6", udpPort)).ping(RCF::Oneway, "asdf");
        Platform::OS::SleepMs(1000);
        BOOST_CHECK(gPings == 2);

        gPings = 0;
        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("255.255.255.255", udpPort)).ping(RCF::Oneway, "asdf");
        Platform::OS::SleepMs(1000);
        BOOST_CHECK(gPings >= 1);
    }

    {
        // server discovery, using an agreed multicast address and port

        X x;
        RCF::RcfServer server( RCF::TcpEndpoint(0));
        server.bind( (I_X *) NULL, x);
        server.start();

        int tcpPort = server.getIpServerTransport().getPort();        

        Broadcast broadcast;
        
        RCF::RcfServer broadcastListener( 
            RCF::UdpEndpoint("0.0.0.0", 0).listenOnMulticast("232.5.5.5"));

        broadcastListener.bind( (I_Broadcast *) NULL, broadcast);
        broadcastListener.start();
        
        int udpPort = broadcastListener.getIpServerTransport().getPort();
        
        RCF::ThreadPtr broadcastThreadPtr( new RCF::Thread( boost::bind(broadcastTask, tcpPort, "232.5.5.5", udpPort, boost::cref(server))));
        server.addJoinFunctor( boost::bind( &RCF::Thread::join, broadcastThreadPtr));
        
        // Wait at most 3 seconds for the broadcast to arrive.
        boost::uint32_t t0 = RCF::getCurrentTimeMs();
        while (!broadcast.mPort && (RCF::getCurrentTimeMs() - t0) < 3*1000)
        {
            Platform::OS::SleepMs(200);
        }

        if (broadcast.mPort)
        {
            RcfClient<I_X> client( RCF::TcpEndpoint(broadcast.mPort));
            client.ping("asdf");
        }
        else
        {
            BOOST_CHECK(1 == 0);
        }
        
    }

    {
        // multiple multicast receivers, sharing a port (pub/sub)
        
        int udpPort = 0;

        std::vector< boost::shared_ptr<RCF::RcfServer> > subscribers(10);
        for (std::size_t i=0; i<subscribers.size(); ++i)
        {
            boost::shared_ptr<RCF::RcfServer> subscriber(
                new RCF::RcfServer( 
                    RCF::UdpEndpoint("0.0.0.0", udpPort)
                        .listenOnMulticast("232.5.5.5")));

            std::auto_ptr<X> px(new X());
            subscriber->bind( (I_X *) NULL, px);
            subscriber->start();

            if (udpPort == 0)
            {
                udpPort = subscriber->getIpServerTransport().getPort();
            }
            
            subscribers[i] = subscriber;
        }


        RcfClient<I_X> publisher( RCF::UdpEndpoint("232.5.5.5", udpPort));
        publisher.getClientStub().setDefaultCallingSemantics(RCF::Oneway);

        gPings = 0;
        publisher.ping("asdf");
        publisher.ping("asdf");
        publisher.ping("asdf");
        Platform::OS::SleepMs(1000);
        BOOST_CHECK(gPings = subscribers.size()*3);
    }

    {
        // multiple broadcast receivers, sharing a port (pub/sub)

        X x1;
        
        RCF::RcfServer server1( 
            RCF::UdpEndpoint("0.0.0.0", 0).enableSharedAddressBinding() );

        server1.bind( (I_X *) NULL, x1);
        server1.start();

        int udpPort = server1.getIpServerTransport().getPort();

        X x2;

        RCF::RcfServer server2( 
            RCF::UdpEndpoint("0.0.0.0", udpPort).enableSharedAddressBinding());

        server2.bind( (I_X *) NULL, x2);
        server2.start();

        gPings = 0;

        RcfClient<I_X>( GCC_3_3_1_HACK RCF::UdpEndpoint("255.255.255.255", udpPort))
            .ping(RCF::Oneway, "asdf");

        Platform::OS::SleepMs(1000);
        BOOST_CHECK(gPings == 2);
    }

    return 0;
}
