
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2007. All rights reserved.
// Consult your license for conditions of use.
// Developed by Jarl Lindrud.
// Contact: jlindrud@hotmail.com .
//******************************************************************************

#include <iostream>

#include <RCF/test/TestMinimal.hpp>
#include <RCF/test/TransportFactories.hpp>

#include <RCF/FilterService.hpp>
#include <RCF/Idl.hpp>
#include <RCF/NamedPipeEndpoint.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/ThreadManager.hpp>

#include <RCF/test/ThreadGroup.hpp>
#include <RCF/test/Usernames.hpp>

#include <RCF/util/Tchar.hpp>

#ifdef BOOST_WINDOWS
#include <RCF/SspiFilter.hpp>
#include <RCF/Win32NamedPipeServerTransport.hpp>
#endif

class Echo
{
public:

    std::string echo(const std::string &s)
    {
        return s;
    }

    std::string echo(const std::string &s, boost::uint32_t waitMs)
    {
        Platform::OS::SleepMs(waitMs);
        return s;
    }

#ifdef BOOST_WINDOWS

    std::string whatsMyWin32UserName()
    {
        // Return clients username, by impersonating and then asking Windows
        // for the name of the currently logged on user.

        RCF::Win32NamedPipeImpersonator impersonator;
        RCF::tstring myUserName = RCF::getMyUserName();
        return util::toString(myUserName);
    }

#else

    std::string whatsMyWin32UserName()
    {
        return "";
    }

#endif

    RCF::ByteBuffer echo(RCF::ByteBuffer s)
    {
        return s;
    }
};

RCF_BEGIN(I_Echo, "I_Echo")
    RCF_METHOD_R1(std::string, echo, const std::string &)
    RCF_METHOD_R2(std::string, echo, const std::string &, boost::uint32_t)
    RCF_METHOD_R0(std::string, whatsMyWin32UserName)
    RCF_METHOD_R1(RCF::ByteBuffer, echo, RCF::ByteBuffer)
RCF_END(I_Echo)

RCF::Mutex gIoMutex;

void clientTask(const RCF::tstring & pipeName, boost::uint32_t waitMs)
{
    RCF::NamedPipeEndpoint ep(pipeName);
    RcfClient<I_Echo> client(ep);

    boost::uint32_t t0 = RCF::getCurrentTimeMs();

    std::string s1 = "asdf";
    std::string s2 = client.echo(s1, waitMs);
    BOOST_CHECK(s1 == s2);

    boost::uint32_t t1 = RCF::getCurrentTimeMs();

    RCF::Lock lock(gIoMutex);
    std::cout << "Waited " << t1-t0 << " ms..." << std::endl;
}

int test_main(int, char**)
{

#ifdef BOOST_WINDOWS
    RCF::tstring pipeName = RCF_T("TestPipe");
#else
    RCF::tstring pipeName = TEMP_DIR "TestPipe";
#endif

    RCF::NamedPipeEndpoint ep(pipeName);

    Echo echo;
    RCF::RcfServer server(ep);
    
    // 10 server threads
    server.setPrimaryThreadCount(10);

    server.bind( (I_Echo *) 0, echo);
    server.start();

    {
        // Make a single call, no waiting

        std::string s1 = "test";
        RCF::NamedPipeEndpoint ep(pipeName);
        RcfClient<I_Echo> client(ep);
        std::string s2 = client.echo(s1);
        BOOST_CHECK( s1 == s2 );
    }

    {
        // Make 5 calls, in parallel, each one waiting 2 seconds

        boost::uint32_t t0 = RCF::getCurrentTimeMs();

        boost::uint32_t waitMs = 2000;
        ThreadGroup clients;
        for (std::size_t i=0; i<5; ++i)
        {
            clients.push_back( RCF::ThreadPtr( new RCF::Thread(
                boost::bind(clientTask, pipeName, waitMs))));
        }
        joinThreadGroup(clients);

        // Total time should be little more than 2 seconds
        boost::uint32_t t1 = RCF::getCurrentTimeMs();
        std::cout << "Total time: " << t1-t0 << " ms..." << std::endl;
        BOOST_CHECK(t1-t0 < 2*waitMs);
    }

#ifdef BOOST_WINDOWS

    {
        // Named pipe impersonation - use our own credentials

        RCF::tstring tUserName = RCF::getMyUserName();
        std::string userName(util::toString(tUserName));
        RCF::NamedPipeEndpoint ep(pipeName);
        RcfClient<I_Echo> client(ep);
        std::string s = client.whatsMyWin32UserName();
        BOOST_CHECK(s == userName);
    }

    {
        // SSPI impersonation - use our own credentials

        // TODO: Win32NamedPipeServerTransport doesn't yet support transport
        // filters.

        if (false)
        {
            RCF::FilterServicePtr filterServicePtr(new RCF::FilterService());
            filterServicePtr->addFilterFactory( 
                RCF::FilterFactoryPtr( new RCF::SspiNtlmFilterFactory()));
            server.addService(filterServicePtr);

            RCF::tstring tUserName = RCF::getMyUserName();
            std::string userName(util::toString(tUserName));
            RCF::NamedPipeEndpoint ep(pipeName);
            RcfClient<I_Echo> client(ep);

            RCF::FilterPtr ntlmFilterPtr( new RCF::SspiNtlmFilter() );
            client.getClientStub().requestTransportFilters(ntlmFilterPtr);

            client.getClientStub().setRemoteCallTimeoutMs(1000*3600);
            std::string s = client.whatsMyWin32UserName();
            BOOST_CHECK(s == userName);

            server.removeService(filterServicePtr);
        }
    }

#endif

    {
        // Test error reporting, by exceeding the max message length.
        RCF::ByteBuffer byteBuffer(50*1024);
        for (std::size_t i=0; i<byteBuffer.getLength(); ++i)
        {
            byteBuffer.getPtr()[i] = i%256 ;
        }
        RCF::NamedPipeEndpoint ep(pipeName);
        RcfClient<I_Echo> client(ep);
        try
        {
            BOOST_CHECK( byteBuffer == client.echo(byteBuffer) );
            BOOST_CHECK(1==0);
        }
        catch(RCF::Exception &e)
        {
            BOOST_CHECK(1==1);

#ifdef BOOST_WINDOWS
            BOOST_CHECK( e.getError() == RCF::RcfError_ClientWriteFail );
#else
            BOOST_CHECK( e.getError() == RCF::RcfError_ServerMessageLength );
#endif

        }

        // This time, turn up the message length and check that it goes through.
        server.getServerTransport().setMaxMessageLength(-1);
        client.getClientStub().getTransport().setMaxMessageLength(-1);
        BOOST_CHECK( byteBuffer == client.echo(byteBuffer) );

    }

    return 0;
}
