
// uncomment to enable VLD leak detection - will automatically link to required libs
//#include "vld.h"
//#include "vldapi.h"

#include <vector>

#include <RCF/test/TestMinimal.hpp>

#include <RCF/Idl.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/PublishingService.hpp>
#include <RCF/SubscriptionService.hpp>
#include <RCF/TcpEndpoint.hpp>
#include <RCF/test/TransportFactories.hpp>
#include <RCF/util/CommandLine.hpp>
#include <RCF/util/Profile.hpp>
#include <RCF/util/Platform/OS/Sleep.hpp>

#include <RCF/test/ThreadGroup.hpp>

// some weird problems with linux, maybe emanating from within asio? Something is causing a SIGABRT signal.
#ifdef __linux__
#include <signal.h>
#include <RCF/util/InitDeinit.hpp>
UTIL_ON_INIT( sigignore(SIGABRT) )
#endif

namespace Test_Notification {

    RCF_BEGIN(I_Events, "I_Events")
        RCF_METHOD_V0(void, onA)
        RCF_METHOD_V1(void, onB, int)
        RCF_METHOD_V2(void, onC, int, const std::string &)
    RCF_END(I_Events)

    class Events
    {
    public:

        Events() : nA(RCF_DEFAULT_INIT), nB(RCF_DEFAULT_INIT), nC(RCF_DEFAULT_INIT)
        {}

        void onA()
        {
            nA++;
        }

        void onB(int n)
        {
            nB++;
        }

        void onC(int n, std::string s)
        {
            nC++;
        }

        int nA;
        int nB;
        int nC;

    };

    class ServerTask
    {
    public:
        ServerTask(
            RCF::PublishingService &publishingService,
            const volatile bool *stopFlag) :
                mPublishingService(publishingService),
                mStopFlag(*stopFlag)
        {}

        void operator()()
        {
            try
            {
                mPublishingService.beginPublish( (I_Events*) 0);
                while (!mStopFlag)
                {
                    mPublishingService.publish( (I_Events*) 0).onA();
                }
                mPublishingService.endPublish( (I_Events*) 0);
            }
            catch(const std::exception &e)
            {
                std::cout << "Server task exception" << std::endl;
                std::cout << RCF::toString(e) << std::endl;
            }
        }

    private:
        RCF::PublishingService &mPublishingService;
        const volatile bool &mStopFlag;
    };

    class ClientTask
    {
    public:
        ClientTask(
            RCF::SubscriptionService &subscriptionService,
            RCF::EndpointPtr endpointPtr) :
                mSubscriptionService(subscriptionService),
                mEndpointPtr(endpointPtr)
        {}

        void operator()()
        {
            Events events;

            // Changed this from 1000 to 100 because it was taking so long on unix platforms.
            for (std::size_t n=0; n<100; ++n)
            {
                try
                {

                    mSubscriptionService.beginSubscribe(
                        (I_Events*) 0,
                        events,
                        *mEndpointPtr);

                    mSubscriptionService.endSubscribe(
                        (I_Events*) 0,
                        events);
                        
                }
                catch(const std::exception &e)
                {
                    std::cout << "Client task exception, n=" << n << std::endl;
                    std::cout << RCF::toString(e) << std::endl;
                }
            }
        }

    private:
        RCF::SubscriptionService &mSubscriptionService;
        RCF::EndpointPtr mEndpointPtr;
    };

    void test1(
        RCF::PublishingService &publishingService,
        RCF::SubscriptionService &subscriptionService,
        RCF::EndpointPtr publisherClientEndpointPtr,
        std::vector<Events> &events)
    {
        std::cout << "Test 1: Few subscribers" << std::endl;

        events.clear();
        events.resize(3);

        for (std::size_t j=0; j<3;++j)
        {
            publishingService.beginPublish( (I_Events*) 0);

            for (std::size_t k=0; k<events.size(); ++k)
            {
                events[k] = Events();
                subscriptionService.beginSubscribe(
                    (I_Events *) 0,
                    events[k],
                    *publisherClientEndpointPtr);
            }

            // give the server time to setup the subscriptions
            Platform::OS::Sleep(1);

            for (std::size_t k=0; k<events.size(); ++k)
            {
                subscriptionService.endSubscribe( (I_Events*) 0, events[k]);
                publishingService.publish( (I_Events*) 0).onA();
                publishingService.publish( (I_Events*) 0).onB(1);
                publishingService.publish( (I_Events*) 0).onC(1, "one");

                // give the subscribers time to receive the notifications
                Platform::OS::Sleep(1);
            }

            publishingService.endPublish( (I_Events*) 0);

            for (std::size_t k=0; k<events.size(); ++k)
            {
                BOOST_CHECK(events[k].nA == k);
                BOOST_CHECK(events[k].nB == k);
                BOOST_CHECK(events[k].nC == k);
            }

        }


    }

    void test2(
        RCF::PublishingService &publishingService,
        RCF::SubscriptionService &subscriptionService,
        RCF::EndpointPtr publisherClientEndpointPtr)
    {
        // intensive subscribing and unsubscribing

        std::cout << "Test 2: Intensive subscribing and unsubscribing" << std::endl;
        
        // now fire off all the threads
        bool serverThreadStopFlag = false;
        Thread serverThread(( ServerTask(
            publishingService,
            &serverThreadStopFlag)));

        Platform::OS::SleepMs(1000);

        ThreadGroup clientThreads;
        for (std::size_t j=0; j<2; ++j)
        {
            clientThreads.push_back( ThreadPtr( new Thread( ClientTask(
                subscriptionService,
                publisherClientEndpointPtr->clone()))));
        }
        joinThreadGroup(clientThreads);

        serverThreadStopFlag = true;
        serverThread.join();
    }

    void onSubscriptionDisconnect(std::vector<bool> &v, std::size_t which)
    {
        RCF_ASSERT(which < v.size());
        RCF_ASSERT(v[which] == false);
        v[which] = true;
    }

    void test3(
        RCF::PublishingService &publishingService,
        RCF::SubscriptionService &subscriptionService,
        RCF::EndpointPtr publisherClientEndpointPtr,
        std::vector<Events> &events)
    {
        std::cout << "Test 3: Disconnect notifications" << std::endl;

        events.clear();
        events.resize(10);
        std::vector<bool> disconnectedEvents(10);

        typedef RCF::SubscriptionPtr SubscriptionPtr;
        SubscriptionPtr subscriptionPtr;

        publishingService.beginPublish( (I_Events*) 0);

        for (std::size_t j=0; j<events.size(); ++j)
        {
            // insertion of following line fixed a bizarre ICE with vc6
            std::vector<bool> vb;

            SubscriptionPtr subscriptionPtr =
                subscriptionService.beginSubscribe(
                    (I_Events*) 0,
                    events[j],
                    *publisherClientEndpointPtr,
                    boost::bind(
                        onSubscriptionDisconnect,
                        boost::ref(disconnectedEvents),
                        j));
        }

        // publish to all subscribers
        for (std::size_t j=0; j<events.size(); ++j)
        {
            BOOST_CHECK(events[j].nA == 0);
            BOOST_CHECK(events[j].nB == 0);
            BOOST_CHECK(events[j].nC == 0);
        }

        Platform::OS::Sleep(1);

        publishingService.publish( (I_Events*) 0).onA();
        publishingService.publish( (I_Events*) 0).onB(1);
        publishingService.publish( (I_Events*) 0).onC(1, "one");

        Platform::OS::Sleep(1);

        for (std::size_t j=0; j<events.size(); ++j)
        {
            BOOST_CHECK(events[j].nA == 1);
            BOOST_CHECK(events[j].nB == 1);
            BOOST_CHECK(events[j].nC == 1);
        }


        // check that polling for disconnections works, explicit disconnect

        subscriptionPtr = subscriptionService.getSubscriptionPtr(
            (I_Events*) 0,
            events[0]);

        BOOST_CHECK( subscriptionPtr.get() && subscriptionPtr->isConnected());

        subscriptionService.endSubscribe( (I_Events*) 0, events[0]);

        subscriptionPtr = subscriptionService.getSubscriptionPtr(
            (I_Events*) 0,
            events[0]);

        BOOST_CHECK(!subscriptionPtr);


        // check that polling for disconnections works, implicit disconnect

        for (std::size_t j=1; j<events.size(); ++j)
        {
            subscriptionPtr =
                subscriptionService.getSubscriptionPtr( (I_Events*) 0, events[j]);
            BOOST_CHECK(subscriptionPtr.get() && subscriptionPtr->isConnected());
        }

        publishingService.endPublish( (I_Events*) 0);
        Platform::OS::Sleep(1);

        for (std::size_t j=1; j<events.size(); ++j)
        {
            subscriptionPtr =
                subscriptionService.getSubscriptionPtr( (I_Events*) 0, events[j]);
            BOOST_CHECK(subscriptionPtr.get() && !subscriptionPtr->isConnected());
        }


        // check that we have disconnect notifications on all but first subscription

        BOOST_CHECK(disconnectedEvents[0] == false);
        for (std::size_t j=1; j<disconnectedEvents.size(); ++j)
        {
            BOOST_CHECK(disconnectedEvents[j] == true);
        }

        // have to call endSubscribe() before events[] goes out of scope!
        for (std::size_t j=1; j<events.size(); ++j)
        {
            subscriptionService.endSubscribe( (I_Events*) 0, events[j]);
        }
    }

    void test4(
        RCF::PublishingService &publishingService,
        RCF::SubscriptionService &subscriptionService,
        RCF::EndpointPtr publisherClientEndpointPtr,
        RCF::ServerTransportPtr publisherServerTransportPtr,
        std::vector<Events> &events)
    {
        std::cout << "Test 4: Many subscribers" << std::endl;
        events.clear();
        events.resize(100);

        for (std::size_t j=0; j<5; ++j)
        {
            publishingService.beginPublish( (I_Events*) 0);

            for (std::size_t k=0; k<events.size(); ++k)
            {
                events[k] = Events();
                subscriptionService.beginSubscribe(
                    (I_Events*) 0,
                    events[k],
                    *publisherClientEndpointPtr);
            }

            {
                std::string transportName = typeid(*publisherServerTransportPtr).name();
                util::Profile profile(transportName + ": time spent on publishing calls");
                for (unsigned int k=0; k<events.size()/2; ++k)
                {
                    subscriptionService.endSubscribe( (I_Events*) 0, events[k]);
                    publishingService.publish( (I_Events*) 0).onA();
                    publishingService.publish( (I_Events*) 0).onB(1);
                    publishingService.publish( (I_Events*) 0).onC(1, "one");
                }
            }

            for (std::size_t k=events.size()/2; k<events.size(); ++k)
            {
                subscriptionService.endSubscribe( (I_Events*) 0, events[k]);
            }

            publishingService.endPublish( (I_Events*) 0);
        }
    }

    void shouldNotBeCalled()
    {
        BOOST_CHECK(1==0);
    }

} // namespace Test_Notification

int test_main(int argc, char **argv)
{

    printTestHeader(__FILE__);

    RCF::RcfInitDeinit rcfInitDeinit;

    using namespace Test_Notification;

    util::CommandLine::getSingleton().parse(argc, argv);

    for (std::size_t i=0; i<RCF::getTransportFactories().size(); ++i)
    {
        RCF::SubscriptionPtr subscriptionPtr1;
        RCF::SubscriptionPtr subscriptionPtr2;
        {
            RCF::TransportFactoryPtr transportFactoryPtr;
            std::pair<RCF::ServerTransportPtr, RCF::ClientTransportAutoPtrPtr> transports;

            transportFactoryPtr = RCF::getTransportFactories()[i];
            transports = transportFactoryPtr->createTransports();
            RCF::ServerTransportPtr publisherServerTransportPtr( transports.first );
            RCF::ClientTransportAutoPtr publisherClientTransportAutoPtr( *transports.second );

            transportFactoryPtr = RCF::getTransportFactories()[i];
            transports = transportFactoryPtr->createNonListeningTransports();
            RCF::ServerTransportPtr subscriberServerTransportPtr( transports.first );

            RCF::RcfServer publisher(publisherServerTransportPtr);
            RCF::RcfServer subscriber(subscriberServerTransportPtr);

            // need both I_ServerTransportEx and I_ServerTransportSessionFull for
            // publish/subscribe functionality
            RCF::I_ServerTransportEx *pServerTransportEx =
                dynamic_cast<RCF::I_ServerTransportEx *>(&publisher.getServerTransport());

            if (NULL == pServerTransportEx)
            {
                continue;
            }

            RCF::writeTransportTypes(
                std::cout,
                *publisherServerTransportPtr,
                *publisherClientTransportAutoPtr);

            RCF::PublishingServicePtr publishingServicePtr(new RCF::PublishingService);
            RCF::PublishingService &publishingService = *publishingServicePtr;
            publisher.addService(publishingServicePtr);

            RCF::SubscriptionServicePtr subscriptionServicePtr(new RCF::SubscriptionService);
            RCF::SubscriptionService &subscriptionService = *subscriptionServicePtr;
            subscriber.addService(subscriptionServicePtr);

            publisher.start();
            subscriber.start();

            // NB: must call endSubscribe<>(object) before object goes out of scope!
            // ie, object must outlive the subscription. Otherwise we get memory
            // corruption from invoking methods on a dead object.

            // TODO: only allow shared_ptr's for subscriptions?

            RCF::EndpointPtr publisherClientEndpointPtr(
                publisherClientTransportAutoPtr->getEndpointPtr());

            // the vec nonsense came about while porting to vc6

            std::vector<Events> vec;

            vec.clear();
            test1(
                publishingService,
                subscriptionService,
                publisherClientEndpointPtr,
                vec);

            vec.clear();
            test2(
                publishingService,
                subscriptionService,
                publisherClientEndpointPtr);

            vec.clear();
            test3(
                publishingService,
                subscriptionService,
                publisherClientEndpointPtr,
                vec);

            vec.clear();
            test4(
                publishingService,
                subscriptionService,
                publisherClientEndpointPtr,
                publisherServerTransportPtr,
                vec);

            // Test that a SubscriptionPtr can be destroyed after the
            // corresponding RcfServer.
            publishingService.beginPublish( (I_Events*) 0);

            Events events[2];

            subscriptionPtr1 = subscriptionService.beginSubscribe(
                (I_Events *) 0,
                events[0], 
                *publisherClientEndpointPtr);

            subscriptionPtr2 = subscriptionService.beginSubscribe(
                (I_Events *) 0,
                events[1], 
                *publisherClientEndpointPtr,
                boost::bind(shouldNotBeCalled));

            subscriber.stop();
            publisher.stop();
        }

        subscriptionPtr1.reset();
        subscriptionPtr2.reset();
    }

    return boost::exit_success;
}






